/*
 * Copyright (c) 2015, NVIDIA CORPORATION. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include "nvidia-drm-conftest.h" /* NV_DRM_ATOMIC_MODESET_AVAILABLE */

#if defined(NV_DRM_ATOMIC_MODESET_AVAILABLE)

#include "nvidia-drm-priv.h"
#include "nvidia-drm-mmap.h"
#include "nvidia-drm-gem.h"

static void nvidia_drm_vma_open(struct vm_area_struct *vma)
{
    struct drm_gem_object *gem = vma->vm_private_data;

    drm_gem_object_reference(gem);
}

static void nvidia_drm_vma_release(struct vm_area_struct *vma)
{
    struct drm_gem_object *gem = vma->vm_private_data;

    drm_gem_object_unreference_unlocked(gem);
}

const struct vm_operations_struct nv_drm_gem_vma_ops = {
    .open  = nvidia_drm_vma_open,
    .close = nvidia_drm_vma_release,
};

int nvidia_drm_gem_mmap(struct file *filp, struct vm_area_struct *vma)
{
    struct drm_file *file_priv = (struct drm_file*)filp->private_data;

    struct drm_device *dev = file_priv->minor->dev;
    struct nvidia_drm_device *nv_dev = dev->dev_private;

    struct drm_gem_object *gem = NULL;
    struct nvidia_drm_gem_object *nv_gem = NULL;
    unsigned long size = vma->vm_end - vma->vm_start;
    unsigned long page_offset;
    int ret = -EINVAL;

    if (!nvidia_drm_modeset_enabled(dev)) {
        goto done;
    }

    if ((ret = drm_gem_mmap(filp, vma)) < 0) {
        NV_DRM_DEV_LOG_ERR(
            nv_dev,
            "drm_gem_mmap failed with error code %d",
            ret);
        goto done;
    }

    gem = vma->vm_private_data;
    nv_gem = DRM_GEM_OBJECT_TO_NV_GEM_OBJECT(gem);

    page_offset =
        (unsigned long)(uintptr_t)nv_gem->u.nvkms_memory.pLinearAddress;
    page_offset = page_offset >> PAGE_SHIFT;

    if ((ret = nvidia_drm_remap_pfn_range(vma,
                                          vma->vm_start,
                                          page_offset, size,
                                          vma->vm_page_prot)) < 0) {
        drm_gem_object_reference(gem);

        NV_DRM_DEV_LOG_ERR(
            nv_dev,
            "remap_pfn_range failed with error code %d",
            ret);
    }

done:

    return ret;
}

#endif
